==> TundereBird-0.28.tar.bz2: TundereBird-0.28/core/Smdn.Net.Imap4.Client/Smdn.Net.Imap4.Client/ImapClient.cs <==
// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Net;

using Smdn.Collections;
using Smdn.Net.Imap4.Client.Session;
using Smdn.Net.Imap4.Client.Transaction;
using Smdn.Net.Imap4.Protocol.Client;
using Smdn.Net.MessageAccessProtocols;

namespace Smdn.Net.Imap4.Client {
  public class ImapClient : IDisposable {
    public ImapClientProfile Profile {
      get { return profile; }
    }

    public int Timeout {
      get { return (session == null) ? profile.Timeout : session.TransactionTimeout; }
      set
      {
        profile.Timeout = value;

        if (session != null)
          session.TransactionTimeout = value;
      }
    }

    public int SendTimeout {
      get { return (session == null) ? profile.SendTimeout : session.SendTimeout; }
      set
      {
        profile.SendTimeout = value;

        if (session != null)
          session.SendTimeout = value;
      }
    }

    public int ReceiveTimeout {
      get { return (session == null) ? profile.ReceiveTimeout : session.ReceiveTimeout; }
      set
      {
        profile.ReceiveTimeout = value;

        if (session != null)
          session.ReceiveTimeout = value;
      }
    }

    public bool IsConnected {
      get { return (session != null && session.State != ImapSessionState.NotConnected); }
    }

    internal ImapSession Session {
      get { CheckConnected(); return session; }
    }

    public ImapCapabilityList ServerCapabilities {
      get { return Session.ServerCapabilities; }
    }

    public IDictionary<string, string> ServerID {
      get { return Session.ServerID; }
    }

    public ImapNamespace ServerNamespace {
      get { return Session.Namespaces; }
    }

    /*
     * construction
     */
    public ImapClient()
      : this(new ImapClientProfile())
    {
    }

    public ImapClient(Uri authority)
      : this(new ImapClientProfile(authority))
    {
    }

    public ImapClient(ImapClientProfile profile)
    {
      if (profile == null)
        throw new ArgumentNullException("profile");

      this.profile = profile;
    }

    public ImapClient(string host, string userName)
      : this()
    {
      profile.Host = host;
      profile.UserName = userName;
    }

    public ImapClient(string host, int port, string userName)
      : this()
    {
      profile.Host = host;
      profile.Port = port;
      profile.UserName = userName;
    }

    public ImapClient(string host, int port, string userName, string authType)
      : this()
    {
      profile.Host = host;
      profile.Port = port;
      profile.UserName = userName;
    }

    public void Connect(string password)
    {
      Connect(password, null);
    }

    public void Connect(string password, UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      Connect(new NetworkCredential(profile.UserName, password), createSslStreamCallback);
    }

    public void Connect(ICredentialsByHost credentials)
    {
      Connect(credentials, null);
    }

    public void Connect(ICredentialsByHost credentials, UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      try {
        profile.SetCredentials(credentials);

        ThrowIfError(ImapSessionCreator.CreateSession(profile,
                                                      createSslStreamCallback ?? ImapConnection.CreateSslStream,
                                                      out session));

        // update server info
        if (session.ServerCapabilities.Has(ImapCapability.Namespace))
          session.Namespace();

        if (session.ServerCapabilities.Has(ImapCapability.ID))
          session.ID(null); // TODO: client ID
      }
      finally {
        profile.SetCredentials(null);
      }
    }

    void IDisposable.Dispose()
    {
      Disconnect(false);
    }

    public void Disconnect()
    {
      Disconnect(false);
    }

    public void Logout()
    {
      Disconnect(true);
    }

    public void Disconnect(bool logout)
    {
      if (session != null)
        session.Disconnect(logout);

      session = null;
    }

    private ImapSession session;
    private ImapClientProfile profile;

    /*
     * operations
     */
    private ImapOpenedMailboxInfo openedMailbox;

    public ImapOpenedMailboxInfo OpenedMailbox {
      get { CheckConnected(); return openedMailbox; }
    }

    /*
     * LIST/LSUB
     */
    public ImapMailboxInfo GetInbox()
    {
      return GetMailbox(ImapMailbox.Inbox, ImapMailboxListOptions.Default);
    }

    public ImapMailboxInfo GetInbox(ImapMailboxListOptions options)
    {
      return GetMailbox(ImapMailbox.Inbox, options);
    }

    public ImapMailboxInfo GetMailbox(string mailboxName)
    {
      return GetMailbox(mailboxName, ImapMailboxListOptions.Default);
    }

    public ImapMailboxInfo GetMailbox(string mailboxName, ImapMailboxListOptions options)
    {
      var mailbox = GetMailboxes(mailboxName, null, false, false, options).FirstOrDefault();

      if (mailbox == null)
        throw new ImapMailboxNotFoundException(mailboxName);
      else
        return mailbox;
    }

    public IEnumerable<ImapMailboxInfo> GetMailboxes()
    {
      return GetMailboxes(ImapMailboxListOptions.Default);
    }

    public IEnumerable<ImapMailboxInfo> GetMailboxes(ImapMailboxListOptions options)
    {
      return GetMailboxes(null, null, true, false, options);
    }

    internal IEnumerable<ImapMailboxInfo> GetMailboxes(ImapMailbox mailbox,
                                                       ImapMailboxListOptions options)
    {
      return GetMailboxes(mailbox.Name, mailbox.HierarchyDelimiter, true, mailbox.IsNonHierarchical, options);
    }

    internal IEnumerable<ImapMailboxInfo> GetMailboxes(string mailboxName,
                                                       string hierarchyDelimiter,
                                                       bool patternMatching,
                                                       bool isNonHierarchical,
                                                       ImapMailboxListOptions options)
    {
      string mailboxNamePattern;

      if (patternMatching) {
        if (mailboxName == null) {
          mailboxNamePattern = string.Empty;
        }
        else {
          if (string.IsNullOrEmpty(hierarchyDelimiter))
            yield break;
          else if (isNonHierarchical)
            yield break;

          mailboxNamePattern = mailboxName + hierarchyDelimiter;
        }

        mailboxNamePattern += ((int)(options & ImapMailboxListOptions.TopLevelOnly) != 0)
          ? "%"
          : "*";
      }
      else {
        mailboxNamePattern = mailboxName ?? string.Empty;
      }

      bool subscribedOnly, listRemote, requestStatus;

      TranslateListOptions(options, out subscribedOnly, out listRemote, out requestStatus);

      ImapMailbox[] mailboxes;

      if (ServerCapabilities.Has(ImapCapability.ListExtended)) {
        var selectionOptions = ImapListSelectionOptions.Empty;
        var returnOptions = ImapListReturnOptions.Children;

        if (subscribedOnly)
          selectionOptions += ImapListSelectionOptions.Subscribed;

        if (listRemote)
          selectionOptions += ImapListSelectionOptions.Remote;

        if (requestStatus && ServerCapabilities.Has(ImapCapability.ListStatus)) {
          returnOptions += ImapListReturnOptions.StatusDataItems(ImapMailboxInfo.GetStatusDataItem(ServerCapabilities));
          requestStatus = false; // no need to request STATUS
        }

        ThrowIfError(Session.ListExtended(mailboxNamePattern,
                                          selectionOptions,
                                          returnOptions,
                                          out mailboxes));
      }
      else if (ServerCapabilities.Has(ImapCapability.MailboxReferrals) &&
               listRemote) {
        if (subscribedOnly)
          ThrowIfError(Session.RLsub(mailboxNamePattern, out mailboxes));
        else
          ThrowIfError(Session.RList(mailboxNamePattern, out mailboxes));
      }
      else {
        if (subscribedOnly)
          ThrowIfError(Session.Lsub(mailboxNamePattern, out mailboxes));
        else
          ThrowIfError(Session.List(mailboxNamePattern, out mailboxes));
      }

      var selectedMailboxName = openedMailbox == null ? null : openedMailbox.FullName;

      foreach (var m in mailboxes) {
        if (requestStatus && !m.IsUnselectable && !string.Equals(m.Name, selectedMailboxName))
          ThrowIfError(Session.Status(m, ImapMailboxInfo.GetStatusDataItem(ServerCapabilities)));

        yield return new ImapMailboxInfo(this, m);
      }
    }

    private static void TranslateListOptions(ImapMailboxListOptions options,
                                             out bool subscribedOnly,
                                             out bool listRemote,
                                             out bool requestStatus)
    {
      subscribedOnly  = 0 != (int)(options & ImapMailboxListOptions.SubscribedOnly);
      listRemote      = 0 != (int)(options & ImapMailboxListOptions.Remote);
      requestStatus   = 0 != (int)(options & ImapMailboxListOptions.RequestStatus);
    }

    /*
     * CREATE
     */
    public ImapMailboxInfo CreateMailbox(string mailboxName)
    {
      return CreateMailbox(mailboxName, false);
    }

    public ImapMailboxInfo CreateMailbox(string mailboxName, bool subscribe)
    {
      ImapMailbox createdMailbox;

      ThrowIfError(Session.Create(GetValidatedMailboxName(mailboxName), out createdMailbox));

      if (subscribe)
        ThrowIfError(Session.Subscribe(createdMailbox));

      return new ImapMailboxInfo(this, createdMailbox);
    }

    /*
     * SELECT/EXAMINE
     */
    public ImapOpenedMailboxInfo OpenInbox()
    {
      return OpenMailbox(ImapMailbox.Inbox, false);
    }

    public ImapOpenedMailboxInfo OpenInbox(bool readOnly)
    {
      return OpenMailbox(ImapMailbox.Inbox, readOnly);
    }

    public ImapOpenedMailboxInfo OpenMailbox(ImapMailboxInfo mailbox)
    {
      return OpenMailbox(mailbox.FullName, false);
    }

    public ImapOpenedMailboxInfo OpenMailbox(ImapMailboxInfo mailbox, bool readOnly)
    {
      if (!mailbox.Exists)
        throw new ImapProtocolViolationException("mailbox is not existent");
      else if (mailbox.IsUnselectable)
        throw new ImapProtocolViolationException("mailbox is not existent");

      return OpenMailbox(mailbox.FullName, readOnly);
    }

    public ImapOpenedMailboxInfo OpenMailbox(string mailboxName)
    {
      return OpenMailbox(mailboxName, false);
    }

    public ImapOpenedMailboxInfo OpenMailbox(string mailboxName, bool readOnly)
    {
      if (ServerCapabilities.Has(ImapCapability.CondStore))
        ThrowIfError(readOnly
                     ? Session.ExamineCondStore(GetValidatedMailboxName(mailboxName))
                     : Session.SelectCondStore(GetValidatedMailboxName(mailboxName)));
      else
        ThrowIfError(readOnly
                     ? Session.Examine(GetValidatedMailboxName(mailboxName))
                     : Session.Select(GetValidatedMailboxName(mailboxName)));

      openedMailbox = new ImapOpenedMailboxInfo(this, Session.SelectedMailbox);

      return openedMailbox;
    }

    /*
     * CLOSE
     */
    public void CloseMailbox()
    {
      if (Session.SelectedMailbox != null)
        ThrowIfError(Session.Close());

      openedMailbox = null;
    }

    public void NoOp()
    {
      ThrowIfError(Session.NoOp());
    }

    /*
     * utilities
     */
    internal static ImapTransactionResult ThrowIfError(ImapTransactionResult result)
    {
      if (result.Succeeded)
        return result;
      else
        throw new ImapErrorResponseException(result);
    }

    internal bool IsCapable(ImapCapability capability)
    {
      return ServerCapabilities.Has(capability);
    }

    internal bool IsSelected(ImapMailboxInfo mailbox)
    {
      bool discard;

      return IsSelected(mailbox, out discard);
    }

    internal bool IsSelected(ImapMailboxInfo mailbox, out bool readOnly)
    {
      var selected = (Session.SelectedMailbox == mailbox.Mailbox);

      if (selected)
        readOnly = Session.SelectedMailbox.ReadOnly;
      else
        readOnly = false;

      return selected;
    }

    internal void ThrowIfIncapable(ImapCapability requiredCapability)
    {
      if (!ServerCapabilities.Has(requiredCapability))
        throw new ImapIncapableException(requiredCapability);
    }

    internal void ThrowIfNotSelected(ImapMailboxInfo mailbox)
    {
      if (!IsSelected(mailbox))
        throw new ImapProtocolViolationException(string.Format("mailbox '{0}' has been closed or another mailbox is opened",
                                                               mailbox.FullName));
    }

    internal static string GetValidatedMailboxName(string mailboxName)
    {
      if (mailboxName == null)
        throw new ArgumentNullException("mailboxName");
      else if (mailboxName.Length == 0)
        throw new ArgumentException("empty mailbox name is not allowed", "mailboxName");
      else if(mailboxName.Contains("%") || mailboxName.Contains("*"))
        throw new ArgumentException("mailbox name with wildcards is not allowed", "mailboxName");

      return mailboxName;
    }

    /*
     * methods for checking and validating internal status
     */
    private void CheckConnected()
    {
      if (!IsConnected)
        throw new ImapProtocolViolationException("not connected or disconnected");
    }

    public override string ToString()
    {
      return string.Format("{{ImapClient: Authority='{0}', IsConnected={1}, OpenedMailbox={2}}}",
                           profile.Authority,
                           IsConnected,
                           openedMailbox);
    }
  }
}

==> TundereBird-0.29.tar.bz2: TundereBird-0.29/core/Smdn.Net.Imap4.Client/Smdn.Net.Imap4.Client/ImapClient.cs <==
// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Net;

using Smdn.Collections;
using Smdn.Net.Imap4.Client.Session;
using Smdn.Net.Imap4.Client.Transaction;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client {
  public class ImapClient : IDisposable {
    internal static bool DefaultSubscription = false;
    internal static bool DefaultSelectReadOnly = false;

    public ImapClientProfile Profile {
      get { return profile; }
    }

    public int Timeout {
      get { return (session == null) ? profile.Timeout : session.TransactionTimeout; }
      set
      {
        profile.Timeout = value;

        if (session != null)
          session.TransactionTimeout = value;
      }
    }

    public int SendTimeout {
      get { return (session == null) ? profile.SendTimeout : session.SendTimeout; }
      set
      {
        profile.SendTimeout = value;

        if (session != null)
          session.SendTimeout = value;
      }
    }

    public int ReceiveTimeout {
      get { return (session == null) ? profile.ReceiveTimeout : session.ReceiveTimeout; }
      set
      {
        profile.ReceiveTimeout = value;

        if (session != null)
          session.ReceiveTimeout = value;
      }
    }

    public bool IsConnected {
      get { return (session != null && session.State != ImapSessionState.NotConnected); }
    }

    internal ImapSession Session {
      get { CheckConnected(); return session; }
    }

    public ImapCapabilityList ServerCapabilities {
      get { return Session.ServerCapabilities; }
    }

    public IDictionary<string, string> ServerID {
      get { return Session.ServerID; }
    }

    public ImapNamespace ServerNamespace {
      get { return Session.Namespaces; }
    }

    /*
     * construction
     */
    public ImapClient()
      : this(new ImapClientProfile())
    {
    }

    public ImapClient(Uri authority)
      : this(new ImapClientProfile(authority))
    {
    }

    public ImapClient(ImapClientProfile profile)
    {
      if (profile == null)
        throw new ArgumentNullException("profile");

      this.profile = profile;
    }

    public ImapClient(string host, string userName)
      : this()
    {
      profile.Host = host;
      profile.UserName = userName;
    }

    public ImapClient(string host, int port, string userName)
      : this()
    {
      profile.Host = host;
      profile.Port = port;
      profile.UserName = userName;
    }

    public ImapClient(string host, int port, string userName, string authType)
      : this()
    {
      profile.Host = host;
      profile.Port = port;
      profile.UserName = userName;
    }

    public void Connect(string password)
    {
      Connect(password, null);
    }

    public void Connect(string password, UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      Connect(new NetworkCredential(profile.UserName, password), createSslStreamCallback);
    }

    public void Connect(ICredentialsByHost credentials)
    {
      Connect(credentials, null);
    }

    public void Connect(ICredentialsByHost credentials, UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      try {
        profile.SetCredentials(credentials);

        ThrowIfError(ImapSessionCreator.CreateSession(profile,
                                                      createSslStreamCallback ?? ImapConnection.CreateSslStream,
                                                      out session));

        // update server info
        if (session.ServerCapabilities.Has(ImapCapability.Namespace))
          session.Namespace();

        if (session.ServerCapabilities.Has(ImapCapability.ID))
          session.ID(null); // TODO: client ID
      }
      finally {
        profile.SetCredentials(null);
      }
    }

    void IDisposable.Dispose()
    {
      Disconnect(false);
    }

    public void Disconnect()
    {
      Disconnect(false);
    }

    public void Logout()
    {
      Disconnect(true);
    }

    public void Disconnect(bool logout)
    {
      if (session != null)
        session.Disconnect(logout);

      session = null;
    }

    private ImapSession session;
    private ImapClientProfile profile;

    /*
     * operations
     */
    private ImapOpenedMailboxInfo openedMailbox;

    public ImapOpenedMailboxInfo OpenedMailbox {
      get { CheckConnected(); return openedMailbox; }
    }

    /*
     * LIST/LSUB
     */
    public ImapMailboxInfo GetInbox()
    {
      return GetMailbox(ImapMailbox.Inbox, ImapMailboxListOptions.Default);
    }

    public ImapMailboxInfo GetInbox(ImapMailboxListOptions options)
    {
      return GetMailbox(ImapMailbox.Inbox, options);
    }

    public ImapMailboxInfo GetMailbox(string mailboxName)
    {
      return GetMailbox(mailboxName, ImapMailboxListOptions.Default);
    }

    public ImapMailboxInfo GetMailbox(string mailboxName, ImapMailboxListOptions options)
    {
      var mailbox = GetMailboxNoException(mailboxName, options);

      if (mailbox == null)
        throw new ImapMailboxNotFoundException(mailboxName);
      else
        return mailbox;
    }

    internal ImapMailboxInfo GetMailboxNoException(string mailboxName, ImapMailboxListOptions options)
    {
      return GetMailboxes(mailboxName, null, false, false, options).FirstOrDefault();
    }

    public ImapMailboxInfo GetOrCreateMailbox(string mailboxName)
    {
      return GetOrCreateMailbox(mailboxName, DefaultSubscription, ImapMailboxListOptions.Default);
    }

    public ImapMailboxInfo GetOrCreateMailbox(string mailboxName, ImapMailboxListOptions options)
    {
      return GetOrCreateMailbox(mailboxName, DefaultSubscription, options);
    }

    public ImapMailboxInfo GetOrCreateMailbox(string mailboxName, bool subscribe)
    {
      return GetOrCreateMailbox(mailboxName, subscribe, ImapMailboxListOptions.Default);
    }

    public ImapMailboxInfo GetOrCreateMailbox(string mailboxName, bool subscribe, ImapMailboxListOptions options)
    {
      var mailbox = GetMailboxNoException(mailboxName, options);

      if (mailbox == null)
        return CreateMailbox(mailboxName, subscribe);
      else
        return mailbox;
    }

    public IEnumerable<ImapMailboxInfo> GetMailboxes()
    {
      return GetMailboxes(ImapMailboxListOptions.Default);
    }

    public IEnumerable<ImapMailboxInfo> GetMailboxes(ImapMailboxListOptions options)
    {
      return GetMailboxes(null, null, true, false, options);
    }

    internal IEnumerable<ImapMailboxInfo> GetMailboxes(ImapMailbox mailbox,
                                                       ImapMailboxListOptions options)
    {
      return GetMailboxes(mailbox.Name,
                          mailbox.HierarchyDelimiter,
                          true,
                          mailbox.Flags.Has(ImapMailboxFlag.NoInferiors), // ignore HasNoChildren
                          options);
    }

    internal IEnumerable<ImapMailboxInfo> GetMailboxes(string mailboxName,
                                                       string hierarchyDelimiter,
                                                       bool patternMatching,
                                                       bool isNonHierarchical,
                                                       ImapMailboxListOptions options)
    {
      string mailboxNamePattern;

      if (patternMatching) {
        if (mailboxName == null) {
          mailboxNamePattern = string.Empty;
        }
        else {
          if (string.IsNullOrEmpty(hierarchyDelimiter))
            yield break;
          else if (isNonHierarchical)
            yield break;

          mailboxNamePattern = mailboxName + hierarchyDelimiter;
        }

        mailboxNamePattern += ((int)(options & ImapMailboxListOptions.TopLevelOnly) != 0)
          ? "%"
          : "*";
      }
      else {
        mailboxNamePattern = mailboxName ?? string.Empty;
      }

      bool subscribedOnly, listRemote, requestStatus;

      TranslateListOptions(options, out subscribedOnly, out listRemote, out requestStatus);

      ImapMailbox[] mailboxes;

      if (ServerCapabilities.Has(ImapCapability.ListExtended)) {
        var selectionOptions = ImapListSelectionOptions.Empty;
        var returnOptions = ImapListReturnOptions.Children;

        if (subscribedOnly)
          selectionOptions += ImapListSelectionOptions.Subscribed;

        if (listRemote)
          selectionOptions += ImapListSelectionOptions.Remote;

        if (requestStatus && ServerCapabilities.Has(ImapCapability.ListStatus)) {
          returnOptions += ImapListReturnOptions.StatusDataItems(ImapMailboxInfo.GetStatusDataItem(ServerCapabilities));
          requestStatus = false; // no need to request STATUS
        }

        ThrowIfError(Session.ListExtended(mailboxNamePattern,
                                          selectionOptions,
                                          returnOptions,
                                          out mailboxes));
      }
      else if (ServerCapabilities.Has(ImapCapability.MailboxReferrals) &&
               listRemote) {
        if (subscribedOnly)
          ThrowIfError(Session.RLsub(mailboxNamePattern, out mailboxes));
        else
          ThrowIfError(Session.RList(mailboxNamePattern, out mailboxes));
      }
      else {
        if (subscribedOnly)
          ThrowIfError(Session.Lsub(mailboxNamePattern, out mailboxes));
        else
          ThrowIfError(Session.List(mailboxNamePattern, out mailboxes));
      }

      var selectedMailboxName = openedMailbox == null ? null : openedMailbox.FullName;

      foreach (var m in mailboxes) {
        if (requestStatus && !m.IsUnselectable && !string.Equals(m.Name, selectedMailboxName))
          ThrowIfError(Session.Status(m, ImapMailboxInfo.GetStatusDataItem(ServerCapabilities)));

        yield return new ImapMailboxInfo(this, m);
      }
    }

    private static void TranslateListOptions(ImapMailboxListOptions options,
                                             out bool subscribedOnly,
                                             out bool listRemote,
                                             out bool requestStatus)
    {
      subscribedOnly  = 0 != (int)(options & ImapMailboxListOptions.SubscribedOnly);
      listRemote      = 0 != (int)(options & ImapMailboxListOptions.Remote);
      requestStatus   = 0 != (int)(options & ImapMailboxListOptions.RequestStatus);
    }

    /*
     * CREATE
     */
    public ImapMailboxInfo CreateMailbox(string mailboxName)
    {
      return CreateMailbox(mailboxName, DefaultSubscription);
    }

    public ImapMailboxInfo CreateMailbox(string mailboxName, bool subscribe)
    {
      ImapMailbox createdMailbox;

      ThrowIfError(Session.Create(GetValidatedMailboxName(mailboxName), out createdMailbox));

      if (subscribe)
        ThrowIfError(Session.Subscribe(createdMailbox));

      // retrieve mailbox flags
      ThrowIfError(Session.List(createdMailbox));

      return new ImapMailboxInfo(this, createdMailbox);
    }

    /*
     * SELECT/EXAMINE
     */
    public ImapOpenedMailboxInfo OpenInbox()
    {
      return OpenMailbox(ImapMailbox.Inbox, DefaultSelectReadOnly);
    }

    public ImapOpenedMailboxInfo OpenInbox(bool readOnly)
    {
      return OpenMailbox(ImapMailbox.Inbox, readOnly);
    }

    public ImapOpenedMailboxInfo OpenMailbox(ImapMailboxInfo mailbox)
    {
      return OpenMailbox(mailbox, DefaultSelectReadOnly);
    }

    public ImapOpenedMailboxInfo OpenMailbox(ImapMailboxInfo mailbox, bool readOnly)
    {
      if (mailbox == null)
        throw new ArgumentNullException("mailbox");
      if (!mailbox.Exists)
        throw new ImapProtocolViolationException("mailbox is not existent");
      else if (mailbox.IsUnselectable)
        throw new ImapProtocolViolationException("mailbox is not existent");

      if (ServerCapabilities.Has(ImapCapability.CondStore))
        ThrowIfError(readOnly
                     ? Session.ExamineCondStore(mailbox.FullName)
                     : Session.SelectCondStore(mailbox.FullName));
      else
        ThrowIfError(readOnly
                     ? Session.Examine(mailbox.FullName)
                     : Session.Select(mailbox.FullName));

      openedMailbox = new ImapOpenedMailboxInfo(this, Session.SelectedMailbox);

      return openedMailbox;
    }

    public ImapOpenedMailboxInfo OpenMailbox(string mailboxName)
    {
      return OpenMailbox(mailboxName, DefaultSelectReadOnly);
    }

    public ImapOpenedMailboxInfo OpenMailbox(string mailboxName, bool readOnly)
    {
      return OpenMailbox(GetMailbox(mailboxName), readOnly);
    }

    public ImapOpenedMailboxInfo OpenMailbox(string mailboxName, ImapMailboxListOptions options)
    {
      return OpenMailbox(mailboxName, options, DefaultSelectReadOnly);
    }

    public ImapOpenedMailboxInfo OpenMailbox(string mailboxName, ImapMailboxListOptions options, bool readOnly)
    {
      return OpenMailbox(GetMailbox(mailboxName, options), readOnly);
    }

    /*
     * CLOSE
     */
    public void CloseMailbox()
    {
      if (Session.SelectedMailbox != null)
        ThrowIfError(Session.Close());

      openedMailbox = null;
    }

    /*
     * NOOP
     */
    public void NoOp()
    {
      ThrowIfError(Session.NoOp());
    }

    /*
     * GETQUOTA
     */
    public ImapQuota GetQuota(string quotaRoot)
    {
      if (quotaRoot == null)
        throw new ArgumentNullException("quotaRoot");

      if (!IsCapable(ImapCapability.Quota))
        return null; // as default

      ImapQuota quota;

      ThrowIfError(Session.GetQuota(quotaRoot, out quota));

      return quota;
    }

    public double GetQuotaUsage(string quotaRoot, string resourceName)
    {
      if (quotaRoot == null)
        throw new ArgumentNullException("quotaRoot");
      if (string.IsNullOrEmpty(resourceName))
        throw new ArgumentException("must be non-empty value", "resourceName");

      var quota = GetQuota(quotaRoot);

      if (quota == null)
        return 0.0; // as default

      foreach (var resource in quota.Resources) {
        if (string.Equals(resource.Name, resourceName, StringComparison.OrdinalIgnoreCase)) {
          if (resource.Limit == 0L)
            return 0.0;
          else
            return (double)resource.Usage / (double)resource.Limit;
        }
      }

      throw new ImapProtocolViolationException(string.Format("no such resource name: '{0}'", resourceName));
    }

    /*
     * utilities
     */
    internal static ImapTransactionResult ThrowIfError(ImapTransactionResult result)
    {
      return ThrowIfError(result, null);
    }

    internal static ImapTransactionResult ThrowIfError(ImapTransactionResult result, Func<ImapResponseCode, bool> throwIfError)
    {
      if (result.Succeeded)
        return result;

      if (throwIfError != null && result.TaggedStatusResponse != null)
        if (!throwIfError(result.TaggedStatusResponse.ResponseText.Code))
          return result;

      throw new ImapErrorResponseException(result);
    }

    internal bool IsCapable(ImapCapability capability)
    {
      return ServerCapabilities.Has(capability);
    }

    internal bool IsSelected(ImapMailboxInfo mailbox)
    {
      bool discard;

      return IsSelected(mailbox, out discard);
    }

    internal bool IsSelected(ImapMailboxInfo mailbox, out bool readOnly)
    {
      var selected = Session.SelectedMailbox != null && string.Equals(Session.SelectedMailbox.Name, mailbox.Mailbox.Name);

      if (selected)
        readOnly = Session.SelectedMailbox.ReadOnly;
      else
        readOnly = false;

      return selected;
    }

    internal void ThrowIfIncapable(ImapCapability requiredCapability)
    {
      if (!ServerCapabilities.Has(requiredCapability))
        throw new ImapIncapableException(requiredCapability);
    }

    internal void ThrowIfNotSelected(ImapMailboxInfo mailbox)
    {
      if (!IsSelected(mailbox))
        throw new ImapMailboxClosedException(mailbox.FullName);
    }

    internal static string GetValidatedMailboxName(string mailboxName)
    {
      if (mailboxName == null)
        throw new ArgumentNullException("mailboxName");
      else if (mailboxName.Length == 0)
        throw new ArgumentException("empty mailbox name is not allowed", "mailboxName");
      else if(mailboxName.Contains("%") || mailboxName.Contains("*"))
        throw new ArgumentException("mailbox name with wildcards is not allowed", "mailboxName");

      return mailboxName;
    }

    /*
     * methods for checking and validating internal status
     */
    private void CheckConnected()
    {
      if (!IsConnected)
        throw new ImapProtocolViolationException("not connected or disconnected");
    }

    public override string ToString()
    {
      return string.Format("{{ImapClient: Authority='{0}', IsConnected={1}, OpenedMailbox={2}}}",
                           profile.Authority,
                           IsConnected,
                           openedMailbox);
    }
  }
}
